@@ -25,18 +25,25 @@
  *   v0.95, Added proper HT support, and SMT/CMP infrastructure. Hope it 
  *   	    works. Added scalaing_available_frequencies support and made 
  *   	    default.  
- *   v0.96  Fix HT detection code to default to 1 if cpuid gives strange
+ *   v0.96, Fix HT detection code to default to 1 if cpuid gives strange
  *          result (fixes centrinos). Cleaned up error reporting in 
  *          read_file() (fixes non-SAF drivers. i.e. pmac).
+ *   v0.97, Better handling of multi-core/HT cpus, and fixed a lot of SMP 
+ *          bugs in general.  Now uses affected_cpus if it's there.  Fixed 
+ *          braindead think-o in the multi-cpu speed change logic.  Random
+ *          other small cleanups/suggestions from debian bugtrack.  Fix
+ *          Segfault on exit on SMP systems.
+ *
  *   
  * * Contributions from:
  * 	Warren Togami <warren@togami.com>
  * 	Michael Schreiter <Michael_Schreiter@gmx.de>
  * 	Wolfgang Tremmel <tremmel@garf.de>
  * 	Hans Ulrich Niedermann <kdev@n-dimensional.de>
+ * 	Bdale Garbee, Vince Weaver
  * 	-- And many others who have sent me patches over the past 3 years,
  * 	   some of whom I've never even acknowledged.  Just know that your 
- * 	   support is appreciated, even if i didn't merge in your ideas.  
+ * 	   support is appreciated, even if I didn't merge in your ideas.  
  * 	   Thank you.
  */
 
@@ -87,14 +94,14 @@ typedef struct cpuinfo {
 	unsigned long *freq_table;
 	int table_size;
 	int threads_per_core;
+	int scalable_unit;
 } cpuinfo_t;
 
 /* 
- * Global pointer to the beginning of all our info.  This is so we can tear 
- * stuff down later in the signal handler.
+ * Global array that contains a pointer to all CPU info. 
  */
 
-cpuinfo_t **cpuinfo;
+cpuinfo_t **all_cpus;
 
 /* idea stolen from procps */
 static char buf[1024];
@@ -134,7 +141,7 @@ time_t start_time = 0;
 #define SYSFS_TREE "/sys/devices/system/cpu/"
 #define SYSFS_SETSPEED "scaling_setspeed"
 
-#define VERSION	"0.96"
+#define VERSION	"0.97"
 
 void help(void)
 {
@@ -240,8 +247,12 @@ int get_stat(cpuinfo_t *cpu)
 
 int change_speed(cpuinfo_t *cpu, enum modes mode)
 {
-	int fd, len, err;
+	int fd, len, err, i;
+	cpuinfo_t *save;
 	char writestr[100];
+
+	if (cpu->cpuid != cpu->scalable_unit) 
+		return 0;
 	
 	if (mode == RAISE) {
 		if ((func == AGGRESSIVE) || (func == LEAPS)) {
@@ -258,7 +269,16 @@ int change_speed(cpuinfo_t *cpu, enum modes mode)
 		}
 	}
 
-	cpu->current_speed = cpu->freq_table[cpu->speed_index];
+	/* 
+	 * We need to set the current speed on all virtual CPUs that fall
+	 * into this CPU's scalable unit.
+	 */
+	save = cpu;
+	for (i = save->cpuid; i < (save->cpuid + save->threads_per_core); i++) {
+		cpu = all_cpus[i];
+		cpu->current_speed = save->freq_table[save->speed_index];
+	}
+	cpu = save;
 
 	pprintf(3,"Setting speed to %d\n", cpu->current_speed);
 
@@ -335,13 +355,15 @@ enum modes inline decide_speed(cpuinfo_t *cpu)
 	
 	pprintf(4,"PCT = %f\n", pct);
 	
-	if ((pct > ((float)highwater/100.0)) && 
+	if ((pct >= ((float)highwater/100.0)) && 
 			(cpu->current_speed != cpu->max_speed)) {
 		/* raise speed to next level */
+		pprintf(6, "got here RAISE\n"); 
 		return RAISE;
-	} else if ((pct < ((float)lowwater/100.0)) && 
+	} else if ((pct <= ((float)lowwater/100.0)) && 
 			(cpu->current_speed != cpu->min_speed)) {
 		/* lower speed */
+		pprintf(6, "got here LOWER\n"); 
 		return LOWER;
 	}
 	
@@ -557,12 +579,23 @@ void terminate(int signum)
 	cpuinfo_t *cpu;
 	
 	ncpus = sysconf(_SC_NPROCESSORS_CONF);
+	if (ncpus < 1) ncpus = 1;
 	
-	for(i = ncpus; i > 0; i--) {
-		cpu = cpuinfo[i-1];
-		/* force ourselves back up to the highest Mhz speed */
+	/* 
+	 * for each cpu, force it back to full speed.
+	 * don't mix this with the below statement.
+	 * 
+	 * 5 minutes ago I convinced myself you couldn't 
+	 * mix these two, now I can't remember why.  
+	 */
+	for(i = 0; i < ncpus; i++) {
+		cpu = all_cpus[i];
 		func = LEAPS;
 		change_speed(cpu, RAISE);
+	}
+
+	for(i = 0; i < ncpus; i++) {
+		cpu = all_cpus[i];
 		/* close the /proc/stat fd */
 		close(cpu->fd);
 		/* deallocate everything */
@@ -571,6 +604,7 @@ void terminate(int signum)
 		free(cpu->reading);
 		free(cpu);
 	}
+	free(all_cpus);
 	time_t duration = time(NULL) - start_time;
 	pprintf(1,"Statistics:\n");
 	pprintf(1,"  %d speed changes in %d seconds\n",
@@ -608,12 +642,21 @@ static inline void cpuid(int op, int *eax, int *ebx, int *ecx, int *edx)
 }
 
 /* 
- * a little bit of black magic to try and detect the number of cores per
+ * A little bit of black magic to try and detect the number of cores per
  * processor.  This will have to be added on to for every architecture, 
  * as we learn how to detect them from userspace.  Note: This method
- * assumes uniform processor/thread ID's.
+ * assumes uniform processor/thread ID's.  First look for affected_cpus, 
+ * if not, then fall back to the cpuid check, or just default to 1. 
  *
- * you can always override this by using the -c command line option to 
+ * By 'thread' in this case, I mean 'entity that is part of one scalable
+ * instance'.  For example, a P4 with hyperthreading has 2 threads in 
+ * one scalable instace.  So does an Athlon X2 dual core, because each
+ * core has to have the same speed.  The new Yonah, on the other hand, may
+ * have two scalable elements, as rumors say you can control both cores' 
+ * speed individually.  Lets hope the speedstep driver populates affected_cpus
+ * correctly...
+ *
+ * You can always override this by using the -c command line option to 
  * specify the number of threads per core.  If you do so, it will do a static
  * mapping, uniform for all real processors in the system.  Actually, so 
  * will this one, because there's no way for me to bind to a processor.
@@ -621,20 +664,69 @@ static inline void cpuid(int op, int *eax, int *ebx, int *ecx, int *edx)
  */
 int determine_threads_per_core(int ncpus)
 {
-#ifdef __i386__
+	char filename[100], *p1;
+	int err, count;
+
+	/* if ncpus is one, we don't care */
+	if (ncpus == 1) return 1;
+	
+	/* 
+	 * First look for the affected_cpus file, and count the 
+	 * number of cpus that supports.  Assume this is true for all
+	 * cpus on the system.
+	 */
+	strncpy(filename, SYSFS_TREE, 30);
+	sprintf(filename, "%scpu0/cpufreq/affected_cpus", filename);
+	
+	/* 
+	 * OK, the funkiest system I can think of right now is
+	 * Sun's Niagara processor, which I think would have 32
+	 * "cpus" in one scalable element.  So make this robust 
+	 * enough to at least handle more than 32 affected cpus 
+	 * at once.
+	 *
+	 * NOTE: I don't even know if Niagara supports scaling, 
+	 * I'm dealing with hypotheticals. 
+	 */
+	
+	count = 1;
+	
+	if ((err = read_file(filename, 0, 1)) == 0) {
+		p1 = buf;
+		err = strtoul(p1, &p1, 5);
+		/* 
+		 * The first cpu should always be 0, so err should be 0
+		 * after the first read.  If its anything else, default to
+		 * one, print a message, and move on. 
+		 */
+		if (err != 0) {
+			pprintf(0, "WARN: cpu0 scaling doesn't affect cpu 0?"
+				       " Assuming 1 thread per core.\n");
+			return 1;
+		}
+		while ((err = strtol(p1, &p1, 5)) != 0)
+			count ++;
+		pprintf(1, "about to return count = %d\n", count);
+		return count;	
+	}
+	pprintf(0,"err=%d", err);
+	
+#ifdef __i386__ 
+	/* Only get here if there's no affected_cpus file. */
 	/* 
 	 * XXXjc fix eventually to run on each processor so you
 	 * can support mixed multi and single-core cpus. Need to know
 	 * how to force ourselves to run on one particular processor.
 	 */
-	int eax,ebx,ecx,edx, num;
+	int eax,ebx,ecx,edx, num=1;
 
 	cpuid(1,&eax,&ebx,&ecx,&edx);
 
 	/* 
 	 * Do we support hyperthreading?
 	 * AMD's dual-core will masquerade as HT, so this should work 
-	 * for them too.
+	 * for them too. (update: it does but this doesn't emulate
+	 * the extra ebx parameter, appearently).
 	 */
 	if(edx & 0x08000000) { 
 		/* 
@@ -662,7 +754,7 @@ int determine_threads_per_core(int ncpus)
 int main(int argc, char **argv)
 {
 	cpuinfo_t *cpu;
-	int ncpus, i, j, err, num_real_cpus, threads_per_core;
+	int ncpus, i, j, err, num_real_cpus, threads_per_core, cpubase;
 	enum modes change, change2;
 
 	/* Parse command line args */
@@ -763,7 +855,7 @@ int main(int argc, char **argv)
 		openlog("powernowd", LOG_AUTHPRIV|LOG_PERROR, LOG_DAEMON);
 	
 	/* My ego's pretty big... */
-	pprintf(0,"PowerNow Daemon v%s, (c) 2003-2005 John Clemens\n", 
+	pprintf(0,"PowerNow Daemon v%s, (c) 2003-2006 John Clemens\n", 
 			VERSION);
 
 	/* are we root?? */
@@ -779,19 +871,17 @@ int main(int argc, char **argv)
 	pprintf(1,"  lowwater:      %4d %%\n", lowwater);
 	pprintf(1,"  highwater:     %4d %%\n", highwater);
 	pprintf(1,"  poll interval: %4d ms\n", poll);
-	
-	ncpus = sysconf(_SC_NPROCESSORS_CONF);
-	
-	/* Malloc, initialise data structs */
-	cpu = malloc(sizeof(cpuinfo_t)*ncpus);
-	
-	if (cpu == (cpuinfo_t *)NULL) {
-		perror("Couldn't malloc cpuinfo");
-		return ENOMEM;
-	}
-	memset(cpu, 0, (sizeof(cpuinfo_t)*ncpus));
-	cpuinfo = &cpu;
 
+	/* 
+	 * This should tell us the number of CPUs that Linux thinks we have,
+	 * or, at least GLIBC
+	 */	
+	ncpus = sysconf(_SC_NPROCESSORS_CONF);
+	if (ncpus < 0) {
+		perror("sysconf could not determine number of cpus, assuming 1\n");
+		ncpus = 1;
+	}
+	
 	if (cores_specified) {
 		if (ncpus < t_per_core) {
 			printf("\nWARNING: bogus # of thread per core, assuming 1\n");
@@ -801,6 +891,8 @@ int main(int argc, char **argv)
 		}
 	} else { 
 		threads_per_core = determine_threads_per_core(ncpus);
+		if (threads_per_core < 0) 
+			threads_per_core = 1;
 	}
 	
 	/* We don't support mixed configs yet */
@@ -809,35 +901,53 @@ int main(int argc, char **argv)
 			ncpus, threads_per_core);
 		printf("WARN: Assuming 1.\n");
 		threads_per_core = 1;
-		//help();
-		//exit(ENOTSUP);
+		/*help();
+		exit(ENOTSUP);*/
 	}
 	
 	num_real_cpus = ncpus/threads_per_core;
-	for (i=0;i<ncpus;i++) {
-		cpu[i].threads_per_core = threads_per_core;
+
+	/* Malloc, initialise data structs */
+	all_cpus = (cpuinfo_t **) malloc(sizeof(cpuinfo_t *)*ncpus);
+	if (all_cpus == (cpuinfo_t **)NULL) {
+		perror("Couldn't malloc all_cpus");
+		return ENOMEM;
 	}
 	
-	pprintf(0,"Found %d cpu%s:  -- %d thread%s (or core%s) per physical cpu\n",
-			ncpus/threads_per_core,
-			(ncpus>1)?"s":"",
-			threads_per_core,	
-			(threads_per_core>1)?"s":"", 
+	for (i=0; i<ncpus; i++) {
+		all_cpus[i] = (cpuinfo_t *)malloc(sizeof(cpuinfo_t));
+		if (all_cpus[i] == (cpuinfo_t *)NULL) {
+			perror("Couldn't malloc all_cpus");
+			return ENOMEM;
+		}
+		memset(all_cpus[i],0,sizeof(cpuinfo_t));
+	}
+	
+	for (i=0;i<ncpus;i++) {
+		all_cpus[i]->threads_per_core = threads_per_core;
+		all_cpus[i]->scalable_unit = (i/threads_per_core)*threads_per_core;
+	}
+	
+	pprintf(0,"Found %d scalable unit%s:  -- %d 'CPU%s' per scalable unit\n",
+			num_real_cpus,
+			(num_real_cpus>1)?"s":"",
+			threads_per_core,
 			(threads_per_core>1)?"s":"");
 	
 	for (i=0;i<ncpus;i++) {
-		if ((err = get_per_cpu_info(&cpu[i], i)) != 0) {
+		cpu = all_cpus[i];
+		if ((err = get_per_cpu_info(cpu, i)) != 0) {
 			printf("\n");
 			goto out;
 		}
 		pprintf(0,"  cpu%d: %dMhz - %dMhz (%d steps)\n", 
-				cpu[i].cpuid,
-				cpu[i].min_speed / 1000, 
-				cpu[i].max_speed / 1000, 
-				cpu[i].table_size);
-		for(j=0;j<cpu[i].table_size; j++) {
+				cpu->cpuid,
+				cpu->min_speed / 1000, 
+				cpu->max_speed / 1000, 
+				cpu->table_size);
+		for(j=0;j<cpu->table_size; j++) {
 			pprintf(1, "     step%d : %ldMhz\n", j+1, 
-					cpu[i].freq_table[j] / 1000);
+					cpu->freq_table[j] / 1000);
 		}
 	}
 	
@@ -853,15 +963,18 @@ int main(int argc, char **argv)
 	/* Now the main program loop */
 	while(1) {
 		usleep(poll*1000);
-		for(i=0; i<num_real_cpus; i+=j) {
+		for(i=0; i<num_real_cpus; i++) {
 			change = LOWER;
+			cpubase = i*threads_per_core;
+			pprintf(6, "i = %d, cpubase = %d, ",i,cpubase);
 			/* handle SMT/CMP here */
-			for (j=0; j<cpu[i].threads_per_core; j++) {
-				change2 = decide_speed(&cpu[i+j]);
+			for (j=0; j<all_cpus[cpubase]->threads_per_core; j++) {
+				change2 = decide_speed(all_cpus[cpubase+j]);
+				pprintf(6, "change = %d, change2 = %d\n",change,change2);
 				if (change2 > change)
 					change = change2;
 			}
-			if (change != SAME) change_speed(&cpu[i], change);
+			if (change != SAME) change_speed(all_cpus[cpubase], change);
 		}
 	}
 
@@ -873,7 +986,8 @@ out:
 	printf(" - That you have the core cpufreq and cpufreq-userspace\n");
 	printf("   modules loaded into your kernel\n");
 	printf(" - That you have the cpufreq driver for your cpu loaded,\n");
-	printf("   and that it works. (check dmesg for errors)\n");
+	printf("   (for example: powernow-k7), and that it works. Check\n"); 
+	printf("   'dmesg' for errors.\n");
 	printf("If all of the above are true, and you still have problems,\n");
 	printf("please email the author: clemej@alum.rpi.edu\n");
 	
